/*#############################################################################
#                                                                             #
#                                 WinService                                  #
#            Copyright (c) 2001 SOLEIL VERT / David TAILLANDIER               #
#                                                                             #
###############################################################################

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################

The full program is into this file.

Revision history :

december 11th 2001 v1.0  David TAILLANDIER, SOLEIL VERT, Initial release.

#############################################################################*/

#define STRICT
#include <windows.h>
#include <tlhelp32.h>
#include <string.h>

#define _(text)  text      /* for futur translations */

/* run_mode values */
#define INVALID_MODE     0
#define LAUNCH_MODE      1
#define MONITOR_MODE     2

/*###########################################################################*/

DWORD monitor_mode( LPDWORD unused );
void launch_mode( void );

void check_win9x( void );
DWORD launch( const char * filename );
void register_service( DWORD pid );
DWORD process_id( const char * filename );
void load_kernel32_dll( void );
void free_kernel32_dll( void );

#define err_exit(text)  {MessageBox(NULL,text,_("WinService error"),MB_ICONEXCLAMATION|MB_OK);ExitProcess(0);}

void get_cmd_line( const char * cmdline );
#define skip_blank(text)   { while( is_blank(text) )    text++; }
#define is_blank(text)     (  text[0]==' '  ||  text[0]=='\t'   )
const char * get_registration( const char * text );
const char * get_run_mode( const char * text );
const char * get_prog_name( const char * text );
const char * get_int( const char * text, int * integer );
void usage( void );

/*###########################################################################*/

HINSTANCE kernel32_dll_handle;

int  registration;
int  run_mode;
char prog_name[250];
char prog_cmdline[250];
int  check_delay;
int  respawn_delay;

/*################################ WinMain ##################################*/
#pragma argsused
int WINAPI
WinMain( HINSTANCE unused1, HINSTANCE unused2, PSTR cmdline, int unused3 )
{
    DWORD dummy;

    check_win9x();   /* exits if not the right version of Windows */
    get_cmd_line( cmdline );

    if ( run_mode == LAUNCH_MODE )
        launch_mode();     /* does the job and then exit */

    CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)monitor_mode, 0, 0, &dummy );

    /* message loop */
    {
        MSG mes;
        BOOL retval;

        while ( 1 ) {
            retval = GetMessage( &mes, (HWND)NULL ,0 ,0 );
            if ( retval == -1 )           // error
                return( -1 );
            if ( retval == FALSE )        // WM_QUIT
                return( mes.wParam );
            DispatchMessage( &mes );
        }
    }

    return 0;    /* never reached */
}

/*############################# launch_mode #################################*/
void
launch_mode( void )
{
    DWORD pid;

    load_kernel32_dll();

    pid = launch( prog_cmdline );
    register_service( pid );

    free_kernel32_dll();
    ExitProcess( 0 );
}

/*########################### monitor_mode ##################################*/
#pragma argsused
DWORD
monitor_mode( LPDWORD unused )
{
    DWORD pid;
    HANDLE h;
    DWORD dw;

    load_kernel32_dll();

    register_service( GetCurrentProcessId() );   /* register ourselve as a service */

    pid = process_id( prog_name );
    if ( !pid )             pid = launch( prog_cmdline );
    if ( registration )     register_service( pid );

    while (1)
    {
        Sleep( check_delay * 1000 );

        h = OpenProcess( PROCESS_QUERY_INFORMATION, FALSE, pid );
        if ( h == NULL )                    err_exit( _("Problem with OpenProcess") );
        if ( !GetExitCodeProcess(h,&dw) )   err_exit( _("Error with GetExitCodeProcess") );
        CloseHandle( h );
        if ( dw == STILL_ACTIVE )           continue;

        /* the process is died, restart the program */
        Sleep( respawn_delay * 1000 );
        pid = launch( prog_cmdline );
        if ( registration )                 register_service( pid );
    }

    return 0;     /* never reached */
}

/*########################## load_kernel32_dll ##############################*/
void
load_kernel32_dll( void )
{
    kernel32_dll_handle = LoadLibrary("KERNEL32.DLL");
    if ( kernel32_dll_handle == NULL )
        err_exit( _("Can't load KERNEL32.DLL") );
}

/*########################## free_kernel32_dll ##############################*/
void
free_kernel32_dll( void )
{
    if ( FreeLibrary(kernel32_dll_handle) == FALSE )
        err_exit( _("Error while unloading kernel32.dll") );
}

/*############################## process_id #################################*/
/* returns 0 if the process is not found. Returns the pid otherwise */
DWORD
process_id( const char * filename )
{
    HANDLE (WINAPI * pCreateToolhelp32Snapshot)(DWORD, DWORD);
    BOOL (WINAPI * pProcess32First)(HANDLE, LPPROCESSENTRY32);
    BOOL (WINAPI * pProcess32Next)(HANDLE, LPPROCESSENTRY32);

    HANDLE snapshot;
    PROCESSENTRY32 pe32 = {0};    /* I don't know what the {0} is for, sorry */

    char * p;

    pCreateToolhelp32Snapshot = (HANDLE(WINAPI*)(DWORD,DWORD)) GetProcAddress(kernel32_dll_handle,"CreateToolhelp32Snapshot");
    if ( pCreateToolhelp32Snapshot == NULL )
        err_exit( _("Can't find CreateToolhelp32Snapshot") );
    pProcess32First = (BOOL(WINAPI*)(HANDLE,LPPROCESSENTRY32)) GetProcAddress(kernel32_dll_handle,"Process32First");
    if ( pProcess32First == NULL )
        err_exit( _("Can't find Process32First") );
    pProcess32Next  = (BOOL(WINAPI*)(HANDLE,LPPROCESSENTRY32)) GetProcAddress(kernel32_dll_handle,"Process32Next");
    if ( pProcess32Next == NULL )
        err_exit( _("Can't find Process32Next") );

    /* take a snapshot of running processes */
    snapshot = pCreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
    if ( snapshot == (HANDLE)(-1) )
        err_exit( _("Unable to take a snapshot of running processes") );

    // Walk the snapshot of processes and for each process, get information to display.
    pe32.dwSize = sizeof(PROCESSENTRY32);   // must be filled out before use
    if ( !pProcess32First(snapshot, &pe32))
        err_exit( _("Couldn't walk the list of processes.") );

    do {
        p = strrchr(pe32.szExeFile,'\\');
        if ( p == NULL )        p = pe32.szExeFile;
        else                    p++;
        if ( stricmp(filename,p) == 0 )   return pe32.th32ProcessID;
     } while ( pProcess32Next(snapshot,&pe32) );

    CloseHandle( snapshot );
    return 0;
}

/*######################### check_win9x #####################################*/
/* exit(0) if not running Windows 95 or 98, else just return to caller       */
void
check_win9x( void )
{
    OSVERSIONINFO osver;

    // Get the current OS version
    osver.dwOSVersionInfoSize = sizeof(osver);
    if ( GetVersionEx(&osver) == 0 )
        err_exit( _("Can't get OS version\nRuns only on Windows 95 / 98") );

    if (osver.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
       return;

    if (osver.dwPlatformId == VER_PLATFORM_WIN32_NT)
        err_exit( _("Can't run on Windows NT\nRuns only on Windows 95 / 98") );

    err_exit( _("Runs only on Windows 95 / 98") );

    return;
}

/*################################## launch #################################*/
DWORD
launch( const char * filename )
{
    PROCESS_INFORMATION pinfo;
    STARTUPINFO sinfo;

    sinfo.cb = sizeof(STARTUPINFO);
    sinfo.lpDesktop = NULL;
    sinfo.lpReserved = NULL;
    sinfo.lpTitle = NULL;
    sinfo.dwFlags = 0;
    sinfo.cbReserved2 = 0;
    sinfo.lpReserved2 = NULL;

    if ( !CreateProcess(NULL,(char *)filename,NULL,NULL,0,0,NULL,NULL,&sinfo,&pinfo) )
        err_exit( _("Can't launch application") );

    return pinfo.dwProcessId;
}

/*############################# register_service ############################*/
void
register_service( DWORD pid )
{
    DWORD (* reg_service)(DWORD, DWORD);
    char bug[1000];    /*
                        * well... a bug with RegisterServiceProcess... It seems
                        * to be a bug in the Borland product I use. Not sure.
                        */

    reg_service = (DWORD(*)(DWORD,DWORD)) GetProcAddress(kernel32_dll_handle,"RegisterServiceProcess");
    if ( reg_service == NULL )
        err_exit( _("Can't find RegisterServiceProcess") );

    if ( !reg_service(pid,1) )
        err_exit( _("Unable to register the service") );

    return;
}

/*############################## get_cmd_line ###############################*/
void
get_cmd_line( const char * cmdline )
{
    skip_blank( cmdline );
    if ( !cmdline[0] )     usage();
    cmdline = get_registration( cmdline );

    skip_blank( cmdline );
    cmdline = get_run_mode( cmdline );

    if ( run_mode==LAUNCH_MODE && registration==1 )
       err_exit( _("'R' and 'L' can't coexist.") );

    skip_blank( cmdline );
    if ( run_mode == MONITOR_MODE )    {

        cmdline = get_int( cmdline, &check_delay );

        skip_blank( cmdline );
        cmdline = get_int( cmdline, &respawn_delay );

        skip_blank( cmdline );
    }

    get_prog_name( cmdline );

    return;
}

/*############################ get_registration #############################*/
const char *
get_registration( const char * text )
{
    registration = 0;

    if ( text[0]=='-' || text[0]=='/' )
        text++;

    if ( text[0]=='s' || text[0]=='S' )    {
        registration = 1;
        text++;
    }
    /* else, this is another option */

    return text;
}

/*############################ get_run_mode #################################*/
const char *
get_run_mode( const char * text )
{
    run_mode = INVALID_MODE;

    if ( text[0]=='-' || text[0]=='/' )
        text++;

    if ( text[0]=='m' || text[0]=='M' )
        run_mode = MONITOR_MODE;
    if ( text[0]=='l' || text[0]=='L' )
        run_mode = LAUNCH_MODE;
    text++;

    if ( run_mode==INVALID_MODE || is_blank(text)==0 )
        err_exit( _("invalid <run-mode>") );

    return text;
}

/*############################### get_int ###################################*/
const char *
get_int( const char * text, int * integer )
{
    if ( !text[0] )
       err_exit( _("<check-delay> and/or <respawn-delay> not found") );

    while ( text[0]>='0' && text[0]<='9' )    {
        (*integer) = ((*integer)*10) + (text[0]-'0');
        if ( (*integer)>999 )
           err_exit( _("<check-delay> and/or <respawn-delay> must be lower than 1000") );
        text++;
    }

    if ( !text[0] )
        return text;

    if ( !is_blank(text) )
        err_exit( _("<check-delay> and/or <respawn-delay> must contain only digits") );

    return text;
}

/*########################### get_prog_name #################################*/
const char *
get_prog_name( const char * text )
{
    char * ptr;
    int i;

    if ( !text[0] )
        usage();

    /* full command-line */
    if ( strlen(text) >= sizeof(prog_cmdline) )
        err_exit( _("Command line too long") );
    strcpy( prog_cmdline, text );

    /* program file-name */
    if ( text[0] == '\"' )              /* quoted filename   */
    {
        ptr = strchr( &text[1], '\"' );
        if ( ptr == NULL )
           err_exit( _("Bad filename.\nClosing quote not found.") );
        strncpy( prog_name, &text[1], ptr-(&text[1]) );
    }
    else                             /* unquoted filename */
    {
        ptr = strchr( text, ' ' );
        if ( ptr == NULL )   /* no more option after filename */
        {
            i = strlen( text );
            if ( i >= sizeof(prog_name)-1 )
                err_exit( _("<prog_name> too long") );
            strcpy( prog_name, text );
            return text+i;
        }
        if ( ptr-text >= sizeof(prog_name) )
            err_exit( _("<prog_name> too long") );
        strncpy( prog_name, text, ptr-text );
    }

    return ptr+1;
}

/*############################## usage ######################################*/
void
usage( void )
{
    err_exit( _(
        "WinService 1.0   Copyright (c) 2001 David TAILLANDIER / SOLEIL VERT\n"
        "\n"
        "This program is free software. You should have received a copy of the GNU General Public License along\n"
        "with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,\n"
        "Boston, MA  02111-1307, USA.\n"
        "\n"
        "usage: winservice -L <command-line>\n"
        "usage: winservice [-S] -M <check-delay> <respawn-delay> <command-line>\n"
        "\n"
        "     -L to 'launch'. The <command-line> is executed once and registered as a service. WinService\n"
        "          terminates immediatly. This mode 'transforms' a normal program into a service.\n"
        "     -M to 'monitor'. WinService monitors the execution of <command-line>. Each time <command-line>\n"
        "          dies, WinService re-launchs it. This mode forces a program to act like a service.\n"
        "     -S to register as a service, because the -M option allows to register or not (and the -L option\n"
        "          always registers as a service).\n"
        "     <check-delay> gives the interval in seconds for WinService to check if <command-line> is running.\n"
        "     <respawn-delay> gives the delay to re-launch <command-line> after WinService finds it to be died.\n"
        "     <command-line> is the program you want to run 'like a service'. You can add any parameters.\n"
        "\n"
        "Example: winservice -L calc.exe              This will run calc.exe as a service. Just an example :-)\n"
        "Example: winservice -L \"c:\\name with spaces\\my prog.exe\"\n"
        "Example: winservice -L notepad.exe \"c:\\my documents\\text.txt\"\n"
        "Example: winservice -M 2 6 rnaapp.exe            This runs rnaapp.exe without registering it as a\n"
        "     service. Every 2 seconds, WinService checks if rnaapp.exe is alive. If rnaapp.exe is dead,\n"
        "     WinService waits 6 seconds then re-launch it.\n"
        "\n"
        "Works only with Windows 95/98 because we needed only those. A service is invisible into ctrl-alt-del\n"
        "task list. A service isn't stopped at user logoff (but stopped when the computer stops). You can run\n"
        "WinService when you want ; the usual way is to use the 'RunServices' entry in the registery.\n"
        "This program was initialy done because the remote access shipped with Windows 95/98 is not a service.\n"
        "If the user log off it is impossible to reach his computer. Using WinService to register 'rnaapp.exe'\n"
        "(the remote access program) partialy solves the problem.\n"
        ) );
}

/*############################### END OF FILE ###############################*/


